//  AppCommandHandler
//
//  This class provides the custom MQ command handling for the QACLMorph application
//
//  Message Identifier:      QAM
//  Message Range:           0200 - 0299
//  Trace area assignment:   None
//  Version:                 1.0.3  Build: 38
//
//  Version History:
//
//  1.0.0		14/10/2010		Base Release
//  1.0.3		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "AppCommandHandler.h"

//  Default constructor
AppCommandHandler::AppCommandHandler(void)
{
}

//  Normal constructor
AppCommandHandler::AppCommandHandler(ExecEnvironment *xeParent)
{
	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = 0;
	szStatus[0] = '\0';
	State = 0;
	//  Clear the current transaction queue descriptor
	tqCurrent = NULL;
}

//  Destructor
AppCommandHandler::~AppCommandHandler(void)
{
}

//  CustomCommandHandler
//
//  This function implements the custom command handling functions for the QACLMorph application.
//
//  Parameters
//
//  1  -  WORD			-  The encoded command 
//  2  -  char *		-  The address of a buffer containing any command options will be set
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//  The handler should process it's own custom commands but it will also be passed all of the system
//  commands before they are processed by the system command handler. The handler can change the command
//  tell the system to ignore it by returning(MQ_COMMAND_NULL) or can optionally perform additional actions
//  and then return the same value and have the system handler process it.
//
WORD AppCommandHandler::CustomCommandHandler(WORD wCommand, char *szOptions, int iThreadID)
{
	int				wNextCommand;							//  Next command to be processed
	int				iOldMaxTrans;							//  Old value for Max Concurrent Transactions
	int				iNewMaxTrans;							//  New value for Max Concurrent Transactions
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	switch(wCommand)
	{
	case MQ_COMMAND_STATUS:
		//  This is the system status command - before it is issued we show some custom messages
		if (tqCurrent == NULL) strcpy_s(szMsg, MAX_MSG, MSG_QAM0202E);
		else sprintf_s(szMsg, MAX_MSG, MSG_QAM0201I, tqCurrent->dwStarted, tqCurrent->dwCompleted, tqCurrent->dwStarted - tqCurrent->dwCompleted, tqCurrent->MaxConcurrent);
		xeLocal->LogMessage(szMsg, iThreadID);
		AddInLogMessageText(szMsg, NOERROR);
		//  Inform if the queue is suspended
		if (tqCurrent->QueueIsSuspended)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0205I, tqCurrent->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			AddInLogMessageText(szMsg, NOERROR);
		}
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;
	case MQ_COMMAND_ABORT:
		//  This is the system abort command - remove the permit from the transaction queue
		if (tqCurrent != NULL)
		{
			//  Remove the permit flag from the current queue
			tqCurrent->LocalPermit = FALSE;
		}
		//  Revoke the global permit
		xeLocal->RevokeGlobalPermit();

		//  Set the kernel into fast drain mode
		xeLocal->SetFastDrainMode();
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;
	case MQ_COMMAND_STOP:
		//  This is the system stop command - remove the permit from the transaction queue		
		if (tqCurrent != NULL)
		{
			//  Remove the permit flag from the current queue
			tqCurrent->LocalPermit = FALSE;
		}
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;
	case MQ_COMMAND_SUSPEND:
		//  This is the system suspend command - we suspend the current transactin queue
		if (tqCurrent != NULL)
		{
			//  Set the suspended flag
			tqCurrent->QueueIsSuspended = TRUE;
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0203I, tqCurrent->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			AddInLogMessageText(szMsg, NOERROR);
		}
		AddInSetStatusText(ADDIN_STATUS_SUSPENDED);
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;
	case MQ_COMMAND_RESUME:
		//  This is the system resume command - we resume processing on the current transactin queue
		if (tqCurrent != NULL)
		{
			//  Unset the suspended flag
			tqCurrent->QueueIsSuspended = FALSE;
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0204I, tqCurrent->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			AddInLogMessageText(szMsg, NOERROR);
		}
		AddInSetStatusText(ADDIN_STATUS_RUNNING);
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;

	case MQ_COMMAND_MAXTRANS:
		//  Change the max number of concurrent transaction that are being handled by the Queue
		iNewMaxTrans = atoi(szOptions);						//  Get the new value to be set
		if (iNewMaxTrans == 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0206E, tqCurrent->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			AddInLogMessageText(szMsg, NOERROR);
		}
		else
		{
			iOldMaxTrans = tqCurrent->MaxConcurrent;
			tqCurrent->MaxConcurrent = iNewMaxTrans;
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0207I, tqCurrent->szQName, iOldMaxTrans, iNewMaxTrans);
			xeLocal->LogMessage(szMsg, iThreadID);
			AddInLogMessageText(szMsg, NOERROR);
		}
		wNextCommand = MQ_COMMAND_NULL;
		break;

	default:
		//  Pass the command on to the system command processor
		wNextCommand = wCommand;							//  Pass the command on to the system processor
		break;
	}
	return wNextCommand;
}

//  ParseCustomCommand
//
//  This function will recognize any custom MQ commands that are passed to the command handler.
//
//  Command               Encoding
//
//  MaxTrans			  MQ_COMMAND_USER+1
//
//  Parameters
//
//  1  -  char *		-  The command buffer that was received
//  2  -  char *		-  The address of a buffer where any command options will be set
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//
WORD AppCommandHandler::ParseCustomCommand(char *szCommand, char *szOptions, int iThreadID)
{
	//  Parse the supplied command into an encoded command and optional parameters
	WORD		wEncCmd;

	char		szTempBuffer[MAX_CMD + 1];
	char		*lpCmdPtr;
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
	char		*pNextToken;
#endif
 
	int			iIndex;

	wEncCmd = MQ_COMMAND_INVALID;
	szOptions[0] = '\0';								//  Clear the parameter buffer
	if (strlen(szCommand) == 0)
	{
		return(MQ_COMMAND_NULL);
	}

	strcpy_s(szTempBuffer, MAX_CMD, szCommand);
	for (iIndex = 0; szTempBuffer[iIndex]; iIndex++)
	{
		szTempBuffer[iIndex] = toupper(szTempBuffer[iIndex]);
	}
	lpCmdPtr = strtok_s(szTempBuffer, " ,", &pNextToken);

	if (strcmp(lpCmdPtr, MQ_CMD_STRING_MAXTRANS) == 0)
		wEncCmd = MQ_COMMAND_MAXTRANS;

	if (strlen(szCommand) > strlen(lpCmdPtr))
	{
		strcpy_s(szOptions, MAX_CMD, szCommand + strlen(lpCmdPtr) + 1);
	}
	return(wEncCmd);
}
