//  AppTransactionHandler
//
//  This class provides the custom Transaction Queue handling for the QACLMorph application
//
//  Message Identifier:      QAM
//  Message Range:           0300 - 0399
//  Trace area assignment:   None
//  Version:                 1.0.3  Build: 38
//
//  Version History:
//
//  1.0.0		14/10/2010		Base Release
//  1.0.3		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "AppTransactionHandler.h"

//  Default constructor
AppTransactionHandler::AppTransactionHandler(void)
{
}

//  Normal constructor
AppTransactionHandler::AppTransactionHandler(MTExecutive *xeParent, int iThreadID)
{
	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;
}

//  Destructor
AppTransactionHandler::~AppTransactionHandler(void)
{
}

//  SerializeTransaction
//
//  This function will :-
//  1)  Mark the transaction document that this object refers to with a completion status
//  2)  Delete the transaction object
//
//  Parameters
//
//  1  -  void *				-  Pointer to the transaction object
//  2  -  DBHANDLE				-  Handle of the transaction queue database
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void AppTransactionHandler::SerializeTransaction(void *txObject, HANDLE hdbQueue, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;										//  Return code from Notes API calls
	NOTEHANDLE			hnTransaction;									//  Handle of the Transaction Note
	ACLMorphRequest		*amrOld;										//  ACL Morph Request
	long				lElapsed;										//  Elapsed Time
	long				lTimeFactor = CLOCKS_PER_SEC;					//  Clock tick conversion factor
	NUMBER				nGeneric;										//  Generic Number
	char				szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	//  Cast the returned object to an ACL Morph Request
	amrOld = (ACLMorphRequest *) txObject;

	//  Open the transaction note
	stAPIRC = NSFNoteOpen(hdbQueue, amrOld->nidTX, 0, &hnTransaction);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QAM0304E, amrOld->TransID, amrOld->nidTX, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete amrOld;
		return;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_QAM0305I, amrOld->TransID, tqCurrent->szQName, amrOld->dwDbsUpdated, amrOld->dwDatabasesProcessed);
	LogTransactionMessage(szMsg, amrOld->hTXLog, iThreadID);

	//  Mark the transaction as completed
	MarkTransactionCompleted(amrOld->nidTX, hnTransaction, tqCurrent, iThreadID);

	//  Append the history log to the transaction
	AddHistory(amrOld->hTXLog, hnTransaction, iThreadID);

	//  Write the processing results to the transaction
	NSFItemSetTime(hnTransaction, ITEM_START_UPDATE, &amrOld->tdUpdateStarted);
	nGeneric = amrOld->dwServersProcessed;
	NSFItemSetNumber(hnTransaction, ITEM_STATS_SERVERS, &nGeneric);
	nGeneric = amrOld->dwDirectoriesProcessed;
	NSFItemSetNumber(hnTransaction, ITEM_STATS_DIRS, &nGeneric);
	nGeneric = amrOld->dwDatabasesProcessed;
	NSFItemSetNumber(hnTransaction, ITEM_STATS_DBS, &nGeneric);
	nGeneric = amrOld->dwDbsUpdated;
	NSFItemSetNumber(hnTransaction, ITEM_STATS_ACLS, &nGeneric);
	lElapsed = (amrOld->ctEnd - amrOld->ctStart) / lTimeFactor;
	nGeneric = lElapsed;
	NSFItemSetNumber(hnTransaction, ITEM_STATS_ELAPSED, &nGeneric);

	//  Delete the ACL Rule Set constructed for this transaction
	delete amrOld->acrTransform;

	//  Dispose of the transaction object
	delete amrOld;

	//  Save and close the transaction
	NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
	NSFNoteClose(hnTransaction);

	return;
}

//  MarshallTransaction
//
//  This function will create the request object from a transaction and set the values needed to
//  dispatch it.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  void **				-  Pointer for the returmed address to the transaction object
//  5  -  void **				-  Pointer for the returned address for the execution object
//  6  -  UINT *				-  Pointer for the returned Attributes for the transaction
//  7  -  int *					-  Pointer for the returned Priority for the transaction
//  8  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  int			-   The transaction disposition
//
//	TX_DISPOSITION_NORMAL - Dispatch the transaction object and mark the transaction In Progress
//	TX_DISPOSITION_ERROR  - Mark the transaction Error
//  TX_DISPOSITION_DELAY  - Mark the transaction delayed
//  TX_DISPOSITION_PANIC  - Critical Error, stop the queue
//
//  API Use Notes:
//
//
//
int AppTransactionHandler::MarshallTransaction(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, void **txObject, void **xxObject, UINT *Attrs, int *Priority, int iThreadID)
{
	STATUS				stAPIRC;										//  Return code fron Notes API calls
	ACLMorphRequest		*amrNew;										//  ACL Morph Request
	DXACLRuleSet		*rsRules = NULL;								//  ACL Rule Set
	WORD				wRetLen;										//  Returned length
	char				*pList;											//  Pointer to the text history list
	char				szScope1[MAX_SERVER + 1];						//  First Scope String
	char				szScope2[MAX_DATABASE + 1];						//  Second Scope String
	char				szRSSpec[MAX_URL + 1];							//  Rule Set Specification
	char				szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	//  Remove the old transaction history log from the transaction
	stAPIRC = NSFItemDelete(hnTX, TRANSDOC_HISTORY, (WORD) strlen(TRANSDOC_HISTORY));
	if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_ITEM_NOT_FOUND))
	{
		xeLocal->LogMessage(MSG_QAM0309E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Create a New ACL Morph Request
	amrNew = new ACLMorphRequest();

	//  Create a new Text List to hold the Transaction Log History
	stAPIRC = ListAllocate(0, 0, FALSE, &amrNew->hTXLog, &pList, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QAM0306E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		amrNew->hTXLog = NULLHANDLE;
	}

	//  UnLock the memory for the list
	OSUnlockObject(amrNew->hTXLog);

	//  Set the passed NoteID into the request - it is needed when the transaction comes back
	amrNew->nidTX = nidTX;

	//  Set the TransactionID in the ACL Morph Request
	wRetLen = NSFItemGetText(hnTX, ITEM_TRANSID, amrNew->TransID, MAXWORD);
	amrNew->TransID[wRetLen] = '\0';

	//  Now get request values from the transaction document
	wRetLen = NSFItemGetText(hnTX, ITEM_SERVER, szScope1, MAXWORD);
	szScope1[wRetLen] = '\0';
	wRetLen = NSFItemGetText(hnTX, ITEM_DATABASE, szScope2, MAXWORD);
	szScope2[wRetLen] = '\0';
	wRetLen = NSFItemGetText(hnTX, ITEM_ACLPATT, szRSSpec, MAXWORD);
	szRSSpec[wRetLen] = '\0';

	//  If the ACL Pattern was not specified
	if (szRSSpec[0] == '\0')
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QAM0301E, amrNew->TransID);
		LogTransactionMessage(szMsg, amrNew->hTXLog, iThreadID);
		//  Append the history log to the transaction
		AddHistory(amrNew->hTXLog, hnTX, iThreadID);
		delete amrNew;
		return TX_DISPOSITION_ERROR;
	}

	wRetLen = NSFItemGetText(hnTX, ITEM_ALLOWRECURSION, szMsg, MAXWORD);
	szMsg[wRetLen] = '\0';
	if (strcmp(szMsg, "1") != 0)
	{
		amrNew->RecurseDirectories = FALSE;
	}
	else
	{
		amrNew->RecurseDirectories = TRUE;
	}
			
	//  Set the Database ACL Morpher in the request
	amrNew->ScannersNeeded = SCANNER_DATABASE;					//  Exit is for the D/B Level
	amrNew->Reporters = DXREP_DATABASE;							//  Reporter is for D/B Exits
	amrNew->dxrDatabase = amReporter;							//  Set the object address

	//  Now attempt to parse the Request scope into the request object
	if (!amrNew->ParseRequest(xeLocal, szScope1, szScope2))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QAM0302E, amrNew->TransID);
		LogTransactionMessage(szMsg, amrNew->hTXLog, iThreadID);
		//  Append the history log to the transaction
		AddHistory(amrNew->hTXLog, hnTX, iThreadID);
		delete amrNew;
		return TX_DISPOSITION_ERROR;
	}

	//  Generate a new ACL Rule Set from the spec provided
	rsRules = rsParser->parseTheseRules(szRSSpec, iThreadID);

	//  If the rules could not be built then fail the transaction
	if (rsRules == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QAM0303E, amrNew->TransID);
		LogTransactionMessage(szMsg, amrNew->hTXLog, iThreadID);
		//  Append the history log to the transaction
		AddHistory(amrNew->hTXLog, hnTX, iThreadID);
		delete amrNew;
		return TX_DISPOSITION_ERROR;
	}

	//  Set the address of the ACL Rule Set in the the transaction
	amrNew->acrTransform = rsRules;

	//  Set up the rest of the information required by the interface
	*txObject = amrNew;									//  Address of the request object
	*xxObject = dxTarget;								//  Processed by the Domino Explorer instance
	if (amrNew->AnchorScope == RSCOPE_DATABASE) *Attrs = RQATTR_REJOIN | RQATTR_CMLSL2;
	else *Attrs = RQATTR_REJOIN | RQATTR_CMLSL1;
	*Priority = 20;
	return TX_DISPOSITION_NORMAL;
}

//  LogTransactionMessage
//
//  This function will append a decorated version of the message to the Transaction History Log.
//  It will also pass the message to the system logger.
//
//  Parameters
//
//  1  -  char *				-  Pointer to the undecorated message text
//  2  -  HANDLE				-  Handle of the Transaction History Text List
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void AppTransactionHandler::LogTransactionMessage(char *szMsg, HANDLE hTXLog, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	if (hTXLog != NULLHANDLE)
	{
		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QAM0307E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);
			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}

//  AddHistory
//
//  This function will append the transaction history log to the transaction document.
//
//  Parameters
//
//  1  -  HANDLE				-  Handle of the Transaction History Text List
//  2  -  NOTEHANDLE			-  Handle of the Transaction Note
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void AppTransactionHandler::AddHistory(HANDLE hTXLog, HANDLE hnTX, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	DWORD			dwListSize;									//  Size of the Transaction History Log list
	void			*pList;										//  Pointer to the list structure
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	if (hTXLog == NULLHANDLE) return;

	pList = OSLockObject(hTXLog);
	dwListSize = (DWORD) ListGetSize(pList, FALSE);
	stAPIRC = NSFItemAppend(hnTX, 0, TRANSDOC_HISTORY, (WORD) strlen(TRANSDOC_HISTORY), TYPE_TEXT_LIST, pList, dwListSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QAM0308E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	OSUnlockObject(hTXLog);
	OSMemFree(hTXLog);

	return;
}

