//  DbACLMorpher
//
//  This class is an extension of the DXReporter class that provides a per database invocation
//  of the ACL coersion provided by the ACL handling objects.
//
//  Version:  1.0.3
//  Build:    38
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

#include "DbACLMorpher.h"

//   Default Constructor
DbACLMorpher::DbACLMorpher(void)
{
}

//  Normal Constructor
DbACLMorpher::DbACLMorpher(MTExecutive *xeParent, int iThreadID)
{
	//  Save the Runtime
	xeLocal = xeParent;
}

//  Destructor
DbACLMorpher::~DbACLMorpher(void)
{
}

//  ReportOnThisDb
//
//  This is the entry point used by the DX subsystem to initiate actions on a per D/B basis.
//  This function will trigger the co-ersion of the ACL for the passed database to a conforming state
//  by using the ACL Rule Set provided.
//
//  Parameters
//
//  1  -  DXRequest *	-  Pointer to DXRequest that is currently being serviced
//  2  -  DBHANDLE		-  Handle of the database
//
//  Returns
//
//  BOOL	-  TRUE if ok to continue, FALSE if an error was encountered
//
//  Notes:
//
BOOL DbACLMorpher::ReportOnThisDb(DXRequest *dxrRQ, HANDLE hdbEntity, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	HANDLE				hACL;									//  Handle to the database ACL
	ACLMorphRequest		*amrCurrent;							//  ACL Morph Request
	DXACLRuleSet		*acrClone;								//  Clone of the ruleset for this D/B
	BOOL				SkipUpdates = FALSE;					//  Debugging Setting
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Cast the DXRequest to the ACLMorphRequest
	amrCurrent = (ACLMorphRequest *) dxrRQ;

	//  If there is no current set of rules to apply then just exit
	if (amrCurrent->acrTransform == NULL)
	{
		dxrRQ->ReturnCode = RETURN_NOERROR;
		return TRUE;
	}

	//  Mark the start time of the update
	OSCurrentTIMEDATE(&amrCurrent->tdUpdateStarted);

	//  Read the ACL of the current database
	stAPIRC = NSFDbReadACL(hdbEntity, &hACL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QAM0404E, dxrRQ->szDirectory, dxrRQ->szDatabase, dxrRQ->szServer);
		LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
	}
	else
	{
		//  Create a local clone of the Rule Set
		acrClone = new DXACLRuleSet(xeLocal, iThreadID);

		amrCurrent->acrTransform->clone(acrClone, iThreadID);
		if (acrClone != NULL)
		{
			//  Coerce the ACL to conform to the Rule Set
			if (acrClone->coerce(hACL, iThreadID))
			{
				// Update the ACL - unless skipping
				if (!SkipUpdates)
				{
					//  Save the updated ACL
					stAPIRC = NSFDbStoreACL(hdbEntity, hACL, 0L, 0);
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_QAM0405E,dxrRQ->szDirectory, dxrRQ->szDatabase, dxrRQ->szServer);
						LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
					}
					else
					{
						//  Show that updates have been applied to the ACL
						sprintf_s(szMsg, MAX_MSG, MSG_QAM0403I, dxrRQ->szDirectory, dxrRQ->szDatabase, dxrRQ->szServer);
						LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
						amrCurrent->dwDbsUpdated++;
					}
				}
				else
				{
					LogTransactionMessage(MSG_QAM0402I, amrCurrent->hTXLog, iThreadID);
				}
			}
			delete acrClone;
			acrClone = NULL;
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_QAM0403I, dxrRQ->szDirectory, dxrRQ->szDatabase, dxrRQ->szServer);
			LogTransactionMessage(szMsg, amrCurrent->hTXLog, iThreadID);
		}
		//  Free the memory associated with the ACL
		OSMemFree(hACL);
	}

	dxrRQ->ReturnCode = RETURN_NOERROR;
	return TRUE;
}

//  GetDescription
//
//  This is the entry point used by the DX subsystem to obtain a descriptive string for the module.
//
//  Parameters
//
//  1  -  char *		-  Pointer to the buffer where the description string will be returned
//
//  Returns
//
//  Notes:
//
void DbACLMorpher::GetDescription(char *szDescription)
{
	//  Show the description of this Reporter
	strcpy_s(szDescription, MAX_DESC, REPORTER_DESC);
	return;
}

//  LogTransactionMessage
//
//  This function will append a decorated version of the message to the Transaction History Log.
//  It will also pass the message to the system logger.
//
//  Parameters
//
//  1  -  char *				-  Pointer to the undecorated message text
//  2  -  HANDLE				-  Handle of the Transaction History Text List
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void DbACLMorpher::LogTransactionMessage(char *szMsg, HANDLE hTXLog, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	if (hTXLog != NULLHANDLE)
	{
		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QAM0407E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);

			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}

