// QACKMorph.cpp : Queue driven ACL Transformer
//
///******************************************************************************************************************
//*																													*
//*   File:       QACLMorph.cpp																						*
//*   Suite:      Domino Admin Utilities																			*
//*   Version:    1.0.5	build: 42																					*
//*   Author:     Ian Tree/HMNL																						*
//*																													*
//*   Copyright 2010 - 2011 Hadleigh Marshall Netherlands b.v.														*
//*******************************************************************************************************************
//*	ACL Transformer																									*
//*																													*
//*	This utility command processor will update the ACL of all datases within the specified source so that they 		*
//*	conform to the specification of the supplied XML ACL Pattern. The application runs as a server add-in and 		*
//* and derives it's transactions from a queue in the supplied control database.									*
//*																													*
//*	USAGE:																											*
//*																													*
//*	Load QACLMorph ControlDatabase [-V|-T[:Area]|-D] [-E][-M:mmm] [-X:nn] [=IniFilename]							*
//*																													*
//*	[ControlDatabase]	- The name of the control database, relative to the server's Data Directory					*
//* [-V]|[-T[:Area]]|[-D] - Set logging level to Verbose, Trace (optional the area to trace) or Debug.				*
//*	-M:mmm				- Optional - Switch that sets the number of threads to use in each morph operation			*
//* -X:nn				- Optional - Max Concurrent Transactions (defaults to 1)									*
//*	-E					- Optional - Switch that echoes the logging messages to the server console					*
//*	=IniFilename		- Optional - Notes ini filename to use														*
//*																													*
//*	NOTES:																											*
//*																													*
//*																													*
//*******************************************************************************************************************
//*																													*
//*   History:																										*
//*																													*
//*	1.0.0  - 10/10/2010   -		Base Release																		*
//* 1.0.1  - 14/10/2010   -		Max Concurrent Transaction added to parms (-X:nn)									*
//* 1.0.2  - 17/05/2011   -		Fixed an error with compulsory rules only											*
//* 1.0.3  - 22/10/2011   -		VS 2011 Updates																		*
//* 1.0.4  - 12/11/2011   -		Fixed an issue with changing ACL Entry types.										*
//* 1.0.4  - 27/11/2011   -		Fixed an issue with wildcard matching.												*
//*																													*
//******************************************************************************************************************/

//  Application Includes
#include "QACLMorph.h"


int main(int argc, char* argv[])
{
	AppRunSettings			*arsLocal;								//  Run settings for this execution
	AppCommandHandler		*achLocal = NULL;						//  Application specific MQ command handler
	AppTransactionHandler	*thLocal = NULL;						//  Application specific Transaction Queue handler
	ThreadManagerPolicy		*tmpInit;								//  Initial Thread Manager Policy
	MTExecutive				*xeLocal;								//  Runtime for this application
#if (defined(_DEBUG) && defined(WIN32))
	Helper					*DebugHelper;							//  Debugging Helper
#endif
	DXACLRuleSetParser		*rsParser = NULL;						//  ACL Rule Set Parser
	DbACLMorpher			*amLocal = NULL;						//  Database ACL Morpher
	DominoExplorer			*dxLocal = NULL;						//  Domino Explorer
	TransactionQueue		*tqACLMorph = NULL;						//  Transaction Queue
	TransactionQueue		*tqReturned;							//  Rejoined queue entity

	int						iRejoinRC;								//  Return code from rejoin calls
	BOOL					bProcessorShouldQuit;					//  Process loop control

	char					szBuildID[MAX_BUILD];					//  Build Info
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Show application is starting
	std::cout << APP_TITLE << " (" << APP_NAME << ") Version: " << APP_VERSION << " is starting." << std::endl;

	//  Validate and build the Run Settings for this execution
	arsLocal = new AppRunSettings(argc, argv);
	if (!arsLocal->AllowExecution)									//  Exit silently if just showing usage
	{
		delete arsLocal;
		return APPRC_NOERROR;
	}

	if (!arsLocal->IsValid)											//  Abort if validation failed
	{
		std::cout << MSG_QAM0101S << std::endl;
		return APPRC_FATAL;
	}

	//  Set the identification in the Run Settings
	strcpy_s(arsLocal->APPName, MAXAPPNAME, APP_NAME);
	strcpy_s(arsLocal->APPTitle, MAXAPPTITLE, APP_TITLE);
	strcpy_s(arsLocal->APPVer, MAXAPPVERSION, APP_VERSION);

	//  Create and initialise a Thread Manager Policy
	tmpInit = new ThreadManagerPolicy();
	tmpInit->TPoolPolicy = TPOOL_POLICY_OBEYMAXMIN | TPOOL_POLICY_PRESTARTTARGET;
	tmpInit->MaxThreads = arsLocal->ThreadCount;								//  Max Thread Count
	tmpInit->MinThreads = arsLocal->ThreadCount;								//  Min Thread Count
	tmpInit->TargetThreads = arsLocal->ThreadCount;								//  Initial target Thread Count
	tmpInit->PendingRQECapacity = PCAP_PERTHREAD * arsLocal->ThreadCount;		//  Pending RQE Pool size
	tmpInit->RejoinRQECapacity = RCAP_PERTHREAD * arsLocal->ThreadCount;		//  Rejoin RQE Pool size
	tmpInit->AsyLogPoolEntries = 200;											//  Log pool size is 200 entries
	
	//  Set the priority policy
	tmpInit->PriorityPolicy = PRIO_POLICY_AGERQS | PRIO_POLICY_PREFBOOST;

	//  Set CMLS Mode and the CMLS contraints
	//  The ACLMorph process does not have any sub-tasks
	tmpInit->TPSchedMode = TPOOL_MODE_CMLS;
	tmpInit->MaxPctL1Threads = 40;												//  40% allocated to scavengers
	tmpInit->MaxPctL2Threads = 60;												//  60% allocated to UOWs		

	//  Initialise the runtime environment
	xeLocal = new MTExecutive(arsLocal, tmpInit, argc, argv);
	if (!xeLocal->IsInitialised)												//  Abort if runtime failed to initialise
	{
		std::cout << MSG_QAM0101S << std::endl;
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
		return APPRC_FATAL;
	}

	//  Also check the the Thread Manager was initialised properly
	if (!xeLocal->MTFunctionsAvailable)
	{
		std::cout << MSG_QAM0106S << std::endl;
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
		return APPRC_FATAL;
	}

#if (defined(_DEBUG) && defined(WIN32))
	//  Construct the Debugging Helper object
	DebugHelper = new Helper(xeLocal, 0);
	DebugHelper->ReportMemoryUsage(FALSE, FALSE, 0);
#endif

	//  Report the Engine Build Information
	xeLocal->GetBuildInfo_s(szBuildID, MAX_BUILD);
	sprintf_s(szMsg, MAX_MSG, MSG_QAM0102I, szBuildID);
	xeLocal->LogVerbose(szMsg, 0);

	//  Construct a new DbACLMorpher
	amLocal = new DbACLMorpher(xeLocal, 0);
	
	//  Now construct a Domino Explorer to drive the whole process
	dxLocal = new DominoExplorer(xeLocal, 1, 0);
	if (!dxLocal->IsStarted)
	{
		xeLocal->LogMessage(MSG_QAM0104E, 0);
		xeLocal->Close();
		delete dxLocal;
		delete achLocal;
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
		return APPRC_FATAL;
	}

	//  Construct an ACL Rule Set Parser - this will be passed to the Transaction Handler for it to use
	rsParser = new DXACLRuleSetParser(xeLocal, 0);

	//  Construct the Application Transaction Handler
	thLocal = new AppTransactionHandler(xeLocal, 0);
	thLocal->amReporter = amLocal;
	thLocal->dxTarget = dxLocal;
	thLocal->rsParser = rsParser;
	

	//  Now construct a new Transaction Queue descriptor for the transaction queue
	tqACLMorph = new TransactionQueue();
	//  Set the D/B handle of the repository as the queue database
	tqACLMorph->hdbQueue = xeLocal->hdbRepository;
	//  Set the Queue Name
	strcpy_s(tqACLMorph->szQName, MAX_ELEMENT, ACMQNAME);
	//  Define the views to use for the sub-queues
	strcpy_s(tqACLMorph->szReadyQName, MAX_ELEMENT, QUEUE_NEW_TRANS_VIEW);
	strcpy_s(tqACLMorph->szInProgressQName, MAX_ELEMENT, QUEUE_IP_TRANS_VIEW);
	strcpy_s(tqACLMorph->szDelayedQName, MAX_ELEMENT, QUEUE_DELAYED_TRANS_VIEW);
	//  Set the options that control the queue behaviours
	tqACLMorph->wQueueProtocols = QPFLAG_DEFAULT;				//  Default Behaviour for startup & delay
	tqACLMorph->MaxConcurrent = arsLocal->MaxCTX;				//  Set initial Max Concurrency
	tqACLMorph->MinDelay = 10 * 60;								//  Delay Queue for minimum of 10 mins
	tqACLMorph->DelayCycleSecs = 30;							//  30 seconds per dispatch loop

	//  Contruct the Application Command Handler and attach it to the Thread Monitor
	achLocal = new AppCommandHandler(xeLocal);
	achLocal->tqCurrent = tqACLMorph;							//  Transaction Queue
	xeLocal->AttachCommandHandler(achLocal);

#if (defined(_DEBUG) && defined(WIN32))
	//  Activate the debug command extensions
	achLocal->DebugHelper = DebugHelper;
	thLocal->DebugHelper = DebugHelper;
#endif

	//  Dispatch the Transaction Queue for execution
	xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
							achLocal,
							thLocal,
							tqACLMorph,
							RQATTR_REJOIN | RQATTR_CMLSL0,
							10,
							0);

	//  Main Processing (or not) Loop
	//  Monitor the command handler and respond to any requested changes in state
	bProcessorShouldQuit = FALSE;
	AddInSetStatusText(ADDIN_STATUS_RUNNING);
	while (!bProcessorShouldQuit)
	{
		//  Check if any terminal condition signals have been posted in the Command Handler
		if (achLocal->State & CH_STATE_QUIT)
		{
			AddInSetStatusText(ADDIN_STATUS_TERM);
			xeLocal->LogMessage(MSG_QAM0107I); 
			AddInLogMessageText(MSG_QAM0107I, NOERROR);
			bProcessorShouldQuit = TRUE;
		}
		else
		{
			if (achLocal->State & (CH_STATE_STOP | CH_STATE_ABORT))
			{
				AddInSetStatusText(ADDIN_STATUS_TERM);
				xeLocal->LogMessage(MSG_QAM0108I); 
				AddInLogMessageText(MSG_QAM0108I, NOERROR);
				bProcessorShouldQuit = TRUE;
			}
			else xeLocal->AppWait(TIMEOUT_EVENT_WAIT);				//  Sleep
		}
	}

	//  Wait for the transaction queue to drain before shutting down
	iRejoinRC = xeLocal->GetRejoinRequest(achLocal, (void **) &tqReturned, 0);
	while (iRejoinRC != RJR_NONE_EXIST)
	{	
		if (iRejoinRC != RJR_RETURNED) xeLocal->AppWait(PXR_BUSY_WAIT);
		iRejoinRC = xeLocal->GetRejoinRequest(achLocal, (void **) &tqReturned, 0);
	}

	sprintf_s(szMsg, MAX_MSG, MSG_QAM0105I, tqACLMorph->szQName);
	xeLocal->LogMessage(szMsg, 0);

	//  Detach the current command handler
	xeLocal->AttachCommandHandler(NULL);

	//  Clean up the processing object
	if (rsParser != NULL) delete rsParser;
	if (tqACLMorph != NULL) delete tqACLMorph;
	if (thLocal != NULL) delete thLocal;
	if (dxLocal != NULL) delete dxLocal;
	if (achLocal != NULL) delete achLocal;

#if (defined(_DEBUG) && defined(WIN32))
	DebugHelper->ReportMemoryUsage(FALSE, FALSE, 0);
	delete DebugHelper;
#endif

	//  Processing is completed - shut down, clean up and exit.
	if (!xeLocal->Close())
	{
		std::cout << MSG_QAM0103S << std::endl;
		//  Cleanup the local objects
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
#if (defined(_DEBUG) && defined(WIN32))
		CheckMemoryLeaks();
#endif
		return APPRC_FATAL;
	}

	//  Cleanup the local objects
	delete xeLocal;
	delete tmpInit;
	delete arsLocal;

	//  Terminate the application
	std::cout << APP_TITLE << " (" << APP_NAME << ") Version: " << APP_VERSION << " has completed normally." << std::endl;
#if (defined(_DEBUG) && defined(WIN32))
	CheckMemoryLeaks();
#endif
	return APPRC_NOERROR;
}

